/*
 *****************************************************************************************
 *                                                                                       *
 *                                                                                       *
 *****************************************************************************************
 */

#include <Arduino.h>
#include "at858d.h" 

/* 
 *
 */
const byte digitsToDrive[3] = {   
                               0b00111111,  // Drive on only digit 0, (LSD).
                               0b10111110,  // Drive on only digit 1.
                               0b01111110   // Drive on only digit 2, (MSD).
                             };                                
/* 
 * There is another option to do the same, faster, but takes more memory.
 *                   Port D pins:      { 0, 1, 2, 3, 4, 5, 6, 7 }. 
 *                   Display segments: { A, E, D, F, P, C, G, B }.                     
 */             
const boolean symbolsLookUp[11][8] = {     
                                       { 1, 1, 1, 1, 0, 1, 0, 1 },  // Symbol '0' coding.
                                       { 0, 0, 0, 0, 0, 1, 0, 1 },  // Symbol '1' coding.
                                       { 1, 1, 1, 0, 0, 0, 1, 1 },  // Symbol '2' coding.
                                       { 1, 0, 1, 0, 0, 1, 1, 1 },  // Symbol '3' coding.
                                       { 0, 0, 0, 1, 0, 1, 1, 1 },  // Symbol '4' coding.
                                       { 1, 0, 1, 1, 0, 1, 1, 0 },  // Symbol '5' coding.
                                       { 1, 1, 1, 1, 0, 1, 1, 0 },  // Symbol '6' coding.
                                       { 1, 0, 0, 0, 0, 1, 0, 1 },  // Symbol '7' coding.
                                       { 1, 1, 1, 1, 0, 1, 1, 1 },  // Symbol '8' coding.
                                       { 1, 0, 0, 1, 0, 1, 1, 1 },  // Symbol '9' coding.
                                       { 0, 0, 0, 0, 0, 0, 1, 0 },  // Symbol '-' coding.
                                     };

byte symbol = 0;
byte sevenSegments = 0b11111111;                                  
byte digit_i = 0;  // Digits index, starting to polling from right side (LSD).
byte segment_i = 0;  // Segments index, starting to polling from segment A.

                       // LSD to MSD.
boolean decimalPoint[3] = { 0, 0, 0 };  // '1' ON, '0' OFF.

                       // LSD to MSD.
byte displayBuffer[3] = { 10, 10, 10 };  // Display buffer.
boolean canPrintDisplay = 1;  // Makeshift, is a flag.

typedef	struct 
{
  unsigned int temp;  // If temp * 1000 'milidegrees' change to long.
  unsigned long uVolts;  // If no GAIN change to int.
} tcTable;

/* 
 * Type K thermocouple, lookup table and linearization.
 * Based on the Miceuz k-thermocouple-lib: https://github.com/Miceuz/k-thermocouple-lib/
 *                           {ºC, uvolts * GAIN}.                                      
 */
const tcTable kTable[29] = {    
                             { 0 , 0 },
                             { 10 , 22629 },
                             { 20 , 45486 },
                             { 30 , 68571 },
                             { 40 , 91884 },
                             { 50 , 115311 },
                             { 60 , 138852 },
                             { 79 , 183825 },
                             { 98 , 228741 },
                             { 116 , 271092 },
                             { 134 , 312987 },
                             { 139 , 324558 },
                             { 155 , 361323 },
                             { 172 , 400197 },
                             { 193 , 447963 },
                             { 212 , 491283 },
                             { 231 , 534831 },
                             { 250 , 578721 },
                             { 269 , 623010 },
                             { 288 , 667584 },
                             { 307 , 712443 },
                             { 326 , 757530 },
                             { 345 , 802788 },
                             { 364 , 848217 },
                             { 383 , 893760 },
                             { 402 , 939474 },
                             { 421 , 985245 },
                             { 440 , 1031187 },
                             { 459 , 1077186 },
                           };

int onChipTemp = 0;
unsigned int kTemp = 0;  // Thermocouple/heater temperature.
unsigned int setTemp = 150;  // From ambient temp. to 459º Celsius.

unsigned int fanSpeed = 100;  

/*
 *****************************************************************************************
 *
 *****************************************************************************************
 */
void setup()
{
  HEATER_OFF;
  DDRB |= _BV(PB1);  // Set PB1 as output for TRIAC control.

  FAN_ON;
  DDRC |= _BV(PC3);  // Set PC3 as output for FAN control.

  DISABLE_DIGITS;
  ALL_SEGS_OFF;  // Set all segments and decimal point to HIGH, or off. 
  DDRB |= (_BV(PB0) | _BV(PB6) | _BV(PB7));  // Set PB0, PB6 and PB7 as output.
  DDRD |= 0xFF;  // Set entire PortD as output (7 segments and decimal point).
  //__asm__("nop\n\t");
  
  PORTB |= (_BV(PB2) | _BV(PB5));  // Set pull-up on for PB2 and PB5.
  DDRB &= ~(_BV(PB2) | _BV(PB5));  // Set PB2 and PB5 as inputs to read push buttons.

  PORTB |= _BV(PB4);  // Set pull-up on for PB4.
  DDRB &= ~_BV(PB4);  // Set as input PB4 to read the internal handle switch.

  // Needs to be called before setup() or some functions won't work.
  init();  // Sets up (timers, ADC, USB, etc).
  
  // Note: Avoids conversions above 15KSPS. Also avoid cross conversions.
  DIDR0 = 0b00111111; // At '1' disable the digital input buffer from ADC5 to ADC0. 
  ADMUX |= ( _BV(REFS1) | _BV(REFS0) );  // Set the ADC reference as internal (1.1V).
  ADCSRA |= ( _BV(ADEN) | _BV(ADPS2) | _BV(ADPS1) ) ;  // Enable ADC, also set 
  ADCSRA &= ~_BV(ADPS1);  // the ADC prescaler to 64 (125KHz).
  // ADCSRB I assume ADC free running mode by default.
  
  setup_timer1();
}

/*
 *****************************************************************************************
 *
 *****************************************************************************************
 */
void loop()
{  
  // Makeshift. Note: A additional delay, will affect the update rate.
  if ( canPrintDisplay )
  {
    display_number( kTemp );  // Update rate at ~42Hz.
    canPrintDisplay = 0;    
  }
}

/*
 *****************************************************************************************
 *
 *****************************************************************************************
 */
void display_number( unsigned int n )  // Change to unsigned int or long.
{  
  displayBuffer[0] = n % 10;
  n = n / 10;
  displayBuffer[1] = n % 10;
  n = n / 10;
  displayBuffer[2] = n % 10;
}

/*
 *****************************************************************************************
 *
 *****************************************************************************************
 */
void setup_timer1( void )
{
  /*
   * ATmega368 running at 8MHz, internal RC oscillator.
   * Prescaler (frequency divider) values:   
   *   CS12   CS11   CS10
   *     0      0      0    Stopped.
   *     0      0      1    /1.  
   *     0      1      0    /8.  
   *     0      1      1    /64.
   *     1      0      0    /256. 
   *     1      0      1    /1024.
   *     1      1      0    External clock on T1 pin, falling edge.
   *     1      1      1    External clock on T1 pin, rising edge.
   */

  uint8_t _sreg = SREG;  // Save SREG.
  cli();  // Disable all interrupts while messing with the register setup.

  // Set prescaler to 256.
  TCCR1B &= ~(_BV(CS10) | _BV(CS11) | _BV(CS12));
  TCCR1B |= _BV(CS12);

  // Set WGM mode 4: CTC using OCR1A.
  TCCR1A &= ~(_BV(WGM10) | _BV(WGM11));
  TCCR1B |= _BV(WGM12);
  TCCR1B &= ~_BV(WGM13);

  // Set normal operation and disconnect PWM pins.
  TCCR1A &= ~(_BV(COM1A1) | _BV(COM1A0) | _BV(COM1B1) | _BV(COM1B0));  

  OCR1A = 31;  // Set top value for TCNT1 ( ~ 1ms ).

  TIMSK1 |= _BV(OCIE1A);  // Enable COMPA isr.

  SREG = _sreg;  // Restore SREG with global interrupt flag.  
}

/*
 *****************************************************************************************
 *
 *****************************************************************************************
 */
ISR( TIMER1_COMPA_vect )
{
  symbol = displayBuffer[digit_i];
  sevenSegments = 0b11111111; 
  if ( symbolsLookUp[symbol][segment_i] )
  {
    sevenSegments &= ~( 1 << segment_i );
  }  

  if ( ( segment_i == 4 ) && decimalPoint[digit_i] )
  {
    sevenSegments = 0b11101111;  // The 5th bit corresponds with decimal point.
  }

  // To avoid display artifacts, keep both port assignments together.
  PORTB = (PORTB | 0xC1) & digitsToDrive[digit_i];  // 0xC1 = 0b11000001.
  PORTD = sevenSegments;
  
  segment_i++;
  
  if ( segment_i == 8 ) 
  {
    segment_i = 0;
    digit_i++;
  }
  
  if ( digit_i == 3 ) 
  {
    digit_i = 0;
    canPrintDisplay = 1;
  }
   
  // In one way or another the cold junction compensation should be added.
  // If the ambient Tª is 24º add 24 to kTemp, or add 54720 = 960uV * GAIN.
  kTemp = kTCuVoltsToTemp( get_uVolts() /* + 54720 */ ) + 24; 
   
  // fanSpeed = getFanSpeed(); 
  
  // WARNING UNDER TEST: Only ON/OFF Control.############################################

  if ( kTemp < setTemp )
  {
    HEATER_ON;
    decimalPoint[0] = 1; 
  }
  else
  {
    HEATER_OFF; 
    decimalPoint[0] = 0;
  }
  
  // WARNING UNDER TEST: Only ON/OFF Control.############################################
}

/*
 *****************************************************************************************
 *
 *****************************************************************************************
 */
int getAtmelTemp( void )
{
  unsigned int wADC;
  int temp;

  /*
   * The internal temperature has to be used with the internal reference, 1.1V.
   * Channel 8 can not be selected with the analogRead function yet.
   */

  // Set the ADC reference as internal (1.1V) and mux to channel 8 (Internal Tª sensor).
  ADMUX = 0b11001000; 

  // Note: Avoids conversions above 15KSPS. Also avoid cross conversions.
  // Wait until the voltage become stable at the S&H after the ADC multiplexer.
  for ( unsigned int i = 0; i < 60; i++ )  // Wait around 2 ADC clock cycles (~16uS).
  { 
    __asm__("nop\n\t");
    
  }

  ADCSRA |= _BV(ADSC);  // Start the ADC.

  // Detect the end of the conversion, take around 13 ADC clock cycles (~100uS).
  while ( bit_is_set( ADCSRA, ADSC ) ); 
  
  wADC = ADCL;       // Reading register "ADCW" takes care of how to read ADCL and ADCH.
  wADC |= ADCH << 8;

  // Applying corrections to reduce measurement errors, and estimate the temperature.
  temp = ( wADC - AT_OFFSET ) * AT_SLOPE;  // And rounding to the nearest integer.  

  return ( temp );  // Return the temperature (Celcius degrees).
}

/*
 *****************************************************************************************
 *
 *****************************************************************************************
 */
unsigned int getFanSpeed( void )
{
  unsigned int wADC;

  // Set the ADC reference as internal (1.1V) and mux to channel 5.
  ADMUX = 0b11000101;   


  // Note: Avoids conversions above 15KSPS. Also avoid cross conversions.
  // Wait until the voltage become stable at the S&H after the ADC multiplexer.
  for ( unsigned int i = 0; i < 60; i++ )  // Wait around 2 ADC clock cycles (~16uS).
  { 
    __asm__("nop\n\t");
    
  }

  ADCSRA |= _BV(ADSC);  // Start the ADC.

  // Detect the end of the conversion, take around 13 ADC clock cycles (~100uS).
  while ( bit_is_set( ADCSRA, ADSC ) );
  
  wADC = ADCL;       // Reading register "ADCW" takes care of how to read ADCL and ADCH.
  wADC |= ADCH << 8;

  return ( wADC );  // Return the temperature (Celcius degrees).
}

/*
 *****************************************************************************************
 *
 *****************************************************************************************
 */
unsigned long get_uVolts( void )
{
  unsigned int wADC;
  unsigned long uVolts;

  // Set the ADC reference as internal (1.1V) and mux to channel 0.
  ADMUX = 0b11000000;  // Move at the beginning of Timer1 ISR? 

  // Note: Avoids conversions above 15KSPS. Also avoid cross conversions.
  // Wait until the voltage become stable at the S&H after the ADC multiplexer.
  for ( unsigned int i = 0; i < 60; i++ )  // Wait around 2 ADC clock cycles (~16uS).
  { 
    __asm__("nop\n\t");
    
  }

  ADCSRA |= _BV(ADSC);  // Start the ADC.

  // Detect the end of the conversion, take around 13 ADC clock cycles (~100uS).
  while ( bit_is_set( ADCSRA, ADSC ) );
  
  wADC = ADCL;       // Reading register "ADCW" takes care of how to read ADCL and ADCH.
  wADC |= ADCH << 8;
 
  uVolts =  ( INT_REF / 1024.0 ) * wADC; // In uV.

  return ( uVolts );   // Return the temperature (Celcius degrees).  
}

/*
 *****************************************************************************************
 * Linear interpolation between two known points (based on the straight line maths).     * 
 *****************************************************************************************
 */
unsigned long interpolate( unsigned long x,
                           unsigned long y0, unsigned long y1, 
                           unsigned long x0, unsigned long x1 ) 
{
  // Operations order: *, /.
  return y0 + ( y1 - y0 ) * ( x - x0 ) / ( x1 - x0 ); 
}

/*
 *****************************************************************************************
 * Returns the index of the first point whose microvolts value is greater than argument. *
 * Rename to searchBigger and add kTable as input parameter?
 *****************************************************************************************
 */
byte search_uVolts( unsigned long uVolts )
{
  boolean notFound = 1;
  byte index = POINTS_COUNT - 1;
  for ( byte i = 0; i < POINTS_COUNT; i++ ) 
  {
    if ( ( kTable[i].uVolts > uVolts ) && ( notFound ) )
    {
      notFound = 0;  // If not found warning, out off range.  
      index = i;
    }
  }
  return index;
}

/*
 *****************************************************************************************
 * Add kTable as input parameter?
 *****************************************************************************************
 */
unsigned long kTCuVoltsToTemp( unsigned long uVolts ) 
{
  byte index = search_uVolts( uVolts );
  
  return interpolate( uVolts,
                      kTable[index-1].temp, kTable[index].temp,
                      kTable[index-1].uVolts, kTable[index].uVolts 
                    );
}
